import pdb
import argparse
import sys
from cmd import Cmd
from pprint import pprint

import pandas as pd
import matplotlib.pyplot as plt


def roundup(x):
    # return int(math.ceil(x / 10.0)) * 10
    return int(round(x))



def evaluate_skill(row):
    '''Determine the skill level of a participant based on the formula
    explained in the paper'''
    skill_map = {
        'Q205_01': 2, # video games
        'Q205_02': 2, # view photos videos
        'Q205_03': 2, # browse the net
        'Q205_05': 2, # word processor, simple docs
        'Q205_04': 5, # email sorting filters
        'Q205_08': 5, # complex docs
        'Q205_06': 10, # assemble from parts
        'Q205_07': 15, # at least one programming language
        }

    sum_points = 0
    for skill, points in skill_map.items():
        if int(row[skill]) == 2:  # 2= checked, 1 = unchecked
            sum_points += points

    if sum_points >= 20:
        return 'expert'
    elif sum_points >= 8:
        return 'medium'
    elif sum_points > 0:
        return 'newbie'
    else:
        return '!!anomaly'







class SurveyAnalyzer():
    def __init__(self, data_path, values_path):
        # load the raw data into a frame, it is tab-separated
        self.data = pd.read_csv(data_path, delimiter='\t', encoding='utf-16-le')

        # get red of the first row, which has the texts of the questions themselves,
        # by moving it into a separate structure
        self.questions = self.data[0:1]
        self.data = self.data[1:]

        self.values = pd.read_csv(values_path, encoding='utf-16-le',
                         delimiter='\t')

        self.augment_data()

    def augment_data(self):
        '''Compute some additional things and add the corresponding columns to
        the dataframe'''
        self.data['SKILLS'] = self.data.apply(lambda row:evaluate_skill(row), axis=1)

        self.data['GENDER'] = self.data.apply(lambda row:self.generic_value_mapper(row, 'Q204'), axis=1)
        self.data['AGE'] = self.data.apply(lambda row:self.generic_value_mapper(row, 'Q203'), axis=1)
        self.data['EDUCATION'] = self.data.apply(lambda row:self.generic_value_mapper(row, 'Q206'), axis=1)
        self.data['LOCATION'] = self.data.apply(lambda row:self.generic_value_mapper(row, 'Q207'), axis=1)

    def generic_value_mapper(self, row, field):
        '''Trasnform numeric values from a row into their word meanings taken
        from the `self.values` data frame.'''
        old_value = int(row[field])  # get the current numeric value of the field
        readable = self.values[self.values.VAR == field]
        # import pdb; pdb.set_trace()
        return readable[readable.RESPONSE == old_value].iloc[0]['MEANING']

    def question_stats(self, question_id):
        '''Produce some basic, human-readable statistics from the data concerning a
        particular question
        :param data:    DataFrame with the survey data set
        :param values:  DataFrame with human-readable values for answer options
        :param question_id: str, question identifier, e.g. "Q204"'''

        # ger the raw answers corresponding to their numeric codes
        numeric_answers = pd.value_counts(self.data[question_id]).to_frame()

        # create a column called `RESPONSE` and use values from the index in it;
        # this will then be used to merge by a shared column with the
        # human-readable values
        numeric_answers['RESPONSE'] = numeric_answers.index

        # rename the first column to express its meaning more accurately
        numeric_answers.columns = ['ANSWER_COUNT', 'RESPONSE']


        # get the strings corresponding to the numeric values in the survey
        readable_values = self.values[self.values.VAR == question_id]
        readable_values = readable_values.set_index('RESPONSE')

        # remove unnecessary column
        del readable_values['VAR']

        # change data type, so the merge is done right
        numeric_answers['RESPONSE'] = numeric_answers['RESPONSE'].apply(pd.to_numeric)


        result = numeric_answers.merge(readable_values, left_on='RESPONSE',
            right_index=True, how='outer')

        result['ANSWER_COUNT'].fillna(0, inplace=True)

        # add another column that will indicate the percentages for each choice
        # import pdb; pdb.set_trace()
        result['PERCENT'] = result['ANSWER_COUNT']/result['ANSWER_COUNT'].sum() * 100
        result['PERCENT'] = result['PERCENT'].apply(roundup)


        del result['RESPONSE']

        question_concise = self.questions[question_id][0]

        return question_concise, result


    def to_latex(self, data):
        '''Render the given data frame as a Latex table, without showing the
        index column'''
        for column in ['ANSWER_COUNT', 'check_count']:
            # I usually don't need these columns in the LaTeX table
            if column in data:
                del data[column]
        return data.sort_values('PERCENT', ascending=False).to_latex(index=False)


    def query(self, question_id):
        if question_id.endswith('_'):
            # question that consists of sub-questions or sub-options. Since I
            # didn't think of an easy way to determine how many sub-questions
            # there are, I just try them all until I reach a KeyError

            # form a new dict where we store our things, and then we will
            # transform it into a pandas dataframe
            answers = {}
            custom_options = None
            for index in range(1, 50):
                # form the sub-question ID, like `Q001_01`
                meta_question_id = '%s%s' % (question_id, str(index).zfill(2))
                try:
                    question, result = sa.question_stats(meta_question_id)
                except KeyError:
                    # we've reached the end of the subquestions.
                    # At this point we can check if there's another column, that
                    # has the same index prefix, but ends with an `a` (e.g. Q001_18a),
                    # where custom-written options are.
                    # WARNING: this column is not necessarily the LAST one! So
                    # we check for it in a more flexible way:
                    # it must START with question_id and end with `a`
                    custom_options_index = [x for x in self.data.columns.tolist() if x.endswith('a') and x.startswith(question_id)]
                    # we always look into the first element, because `custom_options`
                    # always contains a SINGLE element. This was established
                    # empirically in one survey, but maybe somewhere else in the
                    # universe this heuristic won't hold ;-)
                    custom_options = [dev for dev in self.data[custom_options_index].dropna().values.tolist()]
                    break
                else:

                    if 'Checked' in result['MEANING'].values:
                        # this is a question with checkmarks, i.e. yes/no values
                        for count, meaning, percentage in result.values:
                            if meaning == 'Checked':
                                shortened_question = question.split(': ')[-1]
                                answers[shortened_question] = (count, percentage)

                    elif 'Satisfied' in result['MEANING'].values:
                        # this is a Likert-scale question
                        shortened_question = question.split(': ')[-1]
                        answers[shortened_question] = result
                        # TODO unfinished
            import pdb; pdb.set_trace()

            # create a new DataFrame from these options, for later use, as it is
            # more flexible than sticking to the dictionary, because we get all
            # the goodies of Pandas
            rows = [[option, results[0], results[1]] for option, results in answers.items()]
            ndf = pd.DataFrame(rows, columns=['option', 'check_count', 'PERCENT'])
            print('Question: %s' % question)
            print(ndf)
            pprint(custom_options)
            return question, ndf#, custom_options

        else:
            # single question, classic approach
            question, result = sa.question_stats(question_id)
            print('Question: %s\n%s' % (question, result))
            return question, result

    def likert(self, question_id):
        # question that consists of sub-questions or sub-options. Since I
        # didn't think of an easy way to determine how many sub-questions
        # there are, I just try them all until I reach a KeyError

        # form a new dict where we store our things, and then we will
        # transform it into a pandas dataframe

        # for convenience, to accept short-hand notation
        if not question_id.endswith('_'):
            question_id += '_'

        answers = {}
        custom_options = None
        for index in range(1, 50):
            # form the sub-question ID, like `Q001_01`
            meta_question_id = '%s%s' % (question_id, str(index).zfill(2))
            try:
                question, result = sa.question_stats(meta_question_id)
            except KeyError:
                # we've reached the end of the subquestions.
                # At this point we can check if there's another column, that
                # has the same index prefix, but ends with an `a` (e.g. Q001_18a),
                # where custom-written options are.
                # WARNING: this column is not necessarily the LAST one! So
                # we check for it in a more flexible way:
                # it must START with question_id and end with `a`
                custom_options_index = [x for x in self.data.columns.tolist() if x.endswith('a') and x.startswith(question_id)]
                # we always look into the first element, because `custom_options`
                # always contains a SINGLE element. This was established
                # empirically in one survey, but maybe somewhere else in the
                # universe this heuristic won't hold ;-)
                custom_options = [dev for dev in self.data[custom_options_index].dropna().values.tolist()]
                break
            else:

                if 'Satisfied' in result['MEANING'].values:
                    # this is a Likert-scale question
                    shortened_question = question.split(': ')[-1]
                    answers[shortened_question] = result
                    # TODO unfinished
        # import pdb; pdb.set_trace()

        # create a new DataFrame from these options, for later use, as it is
        # more flexible than sticking to the dictionary, because we get all
        # the goodies of Pandas
        # rows = [[option, results[0], results[1]] for option, results in answers.items()]
        # ndf = pd.DataFrame(rows, columns=['option', 'check_count', 'PERCENT'])
        # print('Question: %s' % question)
        # print(ndf)
        # pprint(custom_options)
        return question, answers#, custom_options


class Prompt(Cmd):
    prompt = 'Type question: '

    def __init__(self, analyzer):
        self.analyzer = analyzer
        super().__init__()

        self.last_result = None

    def do_likert(self, line):
        try:
            question, answers = self.analyzer.likert(line.upper())
        except KeyError:
            print('Could not process question')
        else:
            print(question)
            for sub_question, sub_answers in answers.items():
                print(sub_question)
                print(sub_answers)

    def do_query(self, line):
        try:
            question, data = self.analyzer.query(line.upper())
        except KeyError:
            print('Could not process question')
        else:
            self.last_result = data
            if args.latex:
                print(sa.to_latex(data))

    def do_quit(self, line):
        sys.exit()

    def do_latex(self, line):
        if self.last_result is not None:
            self.analyzer.to_latex(self.last_result)
        else:
            print('Nothing to transform to a LaTeX table yet')

    def do_plot(self, line):
        if self.last_result is not None:
            self.last_result.plot()
            plt.show()
        else:
            print('Nothing to plot yet')

    def do_EOF(self, line):
        return True

    def do_debug(self, line):
        self.do_query(line)
        df = self.last_result
        print('Variable `df` contains last DataFrame')
        pdb.set_trace()



if __name__ == '__main__':
    parser = argparse.ArgumentParser(description='Process Soscisurvey data.')
    parser.add_argument('--raw', type=str, help='path to answers CSV file',
                        default='data/data_iot-usability_latest.csv')
    parser.add_argument('--vals', type=str,
                        help='path to CSV file containing value explanations ',
                        default='data/values-misc.csv')
    parser.add_argument('--question', '-q', type=str,
                        help='Question identifier, e.g. Q207')
    parser.add_argument('--latex', dest='latex', action='store_true',
                        help='Also print table in LaTeX format too')

    parser.add_argument('--interactive', dest='interactive', action='store_true',
                        help='Provide an interactive prompt for queries')

    args = parser.parse_args()
    sa = SurveyAnalyzer(args.raw, args.vals)


    if not args.interactive:
        # classic mode, where you run one query, get the result and quit
        if not args.question:
            args.question = input('Type question id: ')

        question, data = sa.query(args.question.upper())

        if args.latex:
            print(sa.to_latex(data))
    else:
        # interactive mode, where the program keeps running and you can run
        # multiple queries
        print('----- Welcome to interactive mode, type `quit` quit.')
        print('----- Syntax `query <question id>`, e.g. `query Q207`')
        questions = [q for q in sa.data.columns.tolist() if q.startswith('Q')]
        print('----- Valid question IDs are: %s' % questions)

        Prompt(sa).cmdloop()

        # while True:
        #     question_id = input('Type question id: ')
        #     try:
        #         question, data = sa.query(question_id.upper())
        #     except KeyError:
        #         print('Could not process question')
        #     else:
        #         print('---------------------------\n\n')
        #         if args.latex:
        #             print(sa.to_latex(data))



    # sa.query('Q204')
    # sa.query('Q007')
    

    # result.plot()
    # plt.show()



'''
-make a new dataframe for
    * questions with _
    * questions with x
'''